import re
from webassets.filter import (
    Filter,
    register_filter
)


class ReplaceFilter(Filter):
    r"""
        A filter that allows arbitrary search/replace of strings using a source
        regex and a replacement string. Unlike cssrewrite this works on strings
        which are not paths and can be used as an output filter.

        Usage:

            replace_static_urls = ReplaceFilter(
                pattern=r'\\s*{{\\s*STATIC_URL\\s*}}\\s*',
                repl=settings.STATIC_URL,
            )
    """

    name = 'replace'
    max_debug_level = None

    def __init__(self, pattern=None, repl=None, as_output=True, **kwargs):
        self.pattern = pattern
        self.repl = repl
        self.as_output = as_output

        super(ReplaceFilter, self).__init__(**kwargs)

    def unique(self):
        """ Return a hashable representation of the parameters to allow different instances of this filter. """
        return self.pattern, self.repl

    def _process(self, _in, out, **kwargs):
        out.write(re.sub(self.pattern, self.repl, _in.read()))

    def output(self, _in, out, **kwargs):
        if self.as_output:
            self._process(_in, out, **kwargs)
        else:
            out.write(_in.read())

    def input(self, _in, out, **kwargs):
        if self.as_output:
            out.write(_in.read())
        else:
            self._process(_in, out, **kwargs)


register_filter(ReplaceFilter)



from webassets.filter import Filter


__all__ = ('Slimit',)


class Slimit(Filter):
    """Minifies JS.

    Requires the ``slimit`` package (https://github.com/rspivak/slimit),
    which is a JavaScript minifier written in Python. It compiles JavaScript
    into more compact code so that it downloads and runs faster.

    It offers mangle and mangle_toplevel options through SLIMIT_MANGLE and SLIMIT_MANGLE_TOPLEVEL
    """

    name = 'slimit'
    options = {"mangle": "SLIMIT_MANGLE", "mangle_toplevel": "SLIMIT_MANGLE_TOPLEVEL"}

    def setup(self):
        try:
            import slimit
        except ImportError:
            raise EnvironmentError('The "slimit" package is not installed.')
        else:
            self.slimit = slimit

    def output(self, _in, out, **kw):
        out.write(self.slimit.minify(_in.read(),
                mangle=self.mangle, mangle_toplevel=self.mangle_toplevel))



import os

from webassets.filter import ExternalTool

__all__ = ('Sass', 'SCSS')


class Sass(ExternalTool):
    """Converts `Sass <http://sass-lang.com/>`_ markup to
    real CSS.

    Requires the Sass executable to be available externally. To install
    it, you might be able to do::

         $ sudo npm install -g sass

    By default, this works as an "input filter", meaning ``sass`` is
    called for each source file in the bundle. This is because the
    path of the source file is required so that @import directives
    within the Sass file can be correctly resolved.

    However, it is possible to use this filter as an "output filter",
    meaning the source files will first be concatenated, and then the
    Sass filter is applied in one go. This can provide a speedup for
    bigger projects.

    To use Sass as an output filter::

        from webassets.filter import get_filter
        sass = get_filter('sass', as_output=True)
        Bundle(...., filters=(sass,))

    However, if you want to use the output filter mode and still also
    use the @import directive in your Sass files, you will need to
    pass along the ``load_paths`` argument, which specifies the path
    to which the imports are relative to (this is implemented by
    changing the working directory before calling the ``sass``
    executable)::

        sass = get_filter('sass', as_output=True, load_paths='/tmp')

    With ``as_output=True``, the resulting concatenation of the Sass
    files is piped to Sass via stdin (``cat ... | sass --stdin ...``)
    and may cause applications to not compile if import statements are
    given as relative paths.

    For example, if a file ``foo/bar/baz.scss`` imports file
    ``foo/bar/bat.scss`` (same directory) and the import is defined as
    ``@import "bat";`` then Sass will fail compiling because Sass
    has naturally no information on where ``baz.scss`` is located on
    disk (since the data was passed via stdin) in order for Sass to
    resolve the location of ``bat.scss``::

        Traceback (most recent call last):
        ...
        webassets.exceptions.FilterError: sass: subprocess had error: stderr=(sass):1: File to import not found or unreadable: bat. (Sass::SyntaxError)
               Load paths:
                 /path/to/project-foo
                on line 1 of standard input
          Use --trace for backtrace.
        , stdout=, returncode=65

    To overcome this issue, the full path must be provided in the
    import statement, ``@import "foo/bar/bat"``, then webassets
    will pass the ``load_paths`` argument (e.g.,
    ``/path/to/project-foo``) to Sass via its ``-I`` flags so Sass can
    resolve the full path to the file to be imported:
    ``/path/to/project-foo/foo/bar/bat``

    Support configuration options:

    SASS_BIN
        The path to the Sass binary. If not set, the filter will
        try to run ``sass`` as if it's in the system path.

    SASS_STYLE
        The style for the output CSS. Can be one of ``expanded`` (default)
        or ``compressed``.

    SASS_AS_OUTPUT
        By default, this works as an "input filter", meaning ``sass`` is
        called for each source file in the bundle. This is because the
        path of the source file is required so that @import directives
        within the Sass file can be correctly resolved.

        However, it is possible to use this filter as an "output filter",
        meaning the source files will first be concatenated, and then the
        Sass filter is applied in one go. This can provide a speedup for
        bigger projects.

        It will also allow you to share variables between files.

    SASS_LOAD_PATHS
        It should be a list of paths relatives to Environment.directory or absolute paths.
        Order matters as sass will pick the first file found in path order.
        These are fed into the -I flag of the sass command and
        is used to control where sass imports code from.
    """
    # TODO: If an output filter could be passed the list of all input
    # files, the filter might be able to do something interesting with
    # it (for example, determine that all source files are in the same
    # directory).

    name = 'sass'
    options = {
        'binary': 'SASS_BIN',
        'use_scss': ('scss', 'SASS_USE_SCSS'),
        'as_output': 'SASS_AS_OUTPUT',
        'load_paths': 'SASS_LOAD_PATHS',
        'style': 'SASS_STYLE',
    }
    max_debug_level = None

    def resolve_path(self, path):
        return self.ctx.resolver.resolve_source(self.ctx, path)

    def _apply_sass(self, _in, out, cd=None):
        # Switch to source file directory if asked, so that this directory
        # is by default on the load path. We could pass it via -I, but then
        # files in the (undefined) wd could shadow the correct files.
        orig_cwd = os.getcwd()
        child_cwd = orig_cwd
        if cd:
            child_cwd = cd

        args = [self.binary or 'sass',
                '--stdin',
                '--style', self.style or 'expanded']

        if not self.use_scss:
            args.append("--indented")

        for path in self.load_paths or []:
            if os.path.isabs(path):
                abs_path = path
            else:
                abs_path = self.resolve_path(path)
            args.extend(['-I', abs_path])

        return self.subprocess(args, out, _in, cwd=child_cwd)

    def input(self, _in, out, source_path, output_path, **kw):
        if self.as_output:
            out.write(_in.read())
        else:
            self._apply_sass(_in, out, os.path.dirname(source_path))

    def output(self, _in, out, **kwargs):
        if not self.as_output:
            out.write(_in.read())
        else:
            self._apply_sass(_in, out)


class SCSS(Sass):
    """Version of the ``sass`` filter that uses the SCSS syntax.
    """

    name = 'scss'

    def __init__(self, *a, **kw):
        assert 'scss' not in kw
        kw['scss'] = True
        super(SCSS, self).__init__(*a, **kw)



from webassets.filter import ExternalTool
from webassets.utils import working_directory


class PostCSS(ExternalTool):
    """Processes CSS code using `PostCSS <http://postcss.org/>`_.

    Requires the ``postcss`` executable to be available externally.
    To install it, you might be able to do::

        $ npm install --global postcss

    You should also install the plugins you want to use::

        $ npm install --global postcss-cssnext

    You can configure postcss in ``postcss.config.js``:

    .. code-block:: javascript

        module.exports = {
            plugins: [
                require('postcss-cssnext')({
                  // optional configuration for cssnext
                })
            ],
        };

    *Supported configuration options*:

    POSTCSS_BIN
        Path to the postcss executable used to compile source files. By
        default, the filter will attempt to run ``postcss`` via the
        system path.

    POSTCSS_EXTRA_ARGS
        Additional command-line options to be passed to ``postcss`` using this
        setting, which expects a list of strings.

    """
    name = 'postcss'

    options = {
        'binary': 'POSTCSS_BIN',
        'extra_args': 'POSTCSS_EXTRA_ARGS',
    }

    max_debug_level = None

    def input(self, in_, out, source_path, **kw):
        # Set working directory to the source file so that includes are found
        args = [self.binary or 'postcss']
        if self.extra_args:
            args.extend(self.extra_args)
        with working_directory(filename=source_path):
            self.subprocess(args, out, in_)



import re
import os

from webassets.test import TempDirHelper, TempEnvironmentHelper


__all__ = ('TempDirHelper', 'TempEnvironmentHelper', 'noop',
           'assert_raises_regex', 'check_warnings', 'normalize_paths')


# Define a noop filter; occasionally in tests we need to define
# a filter to be able to test a certain piece of functionality,.
noop = lambda _in, out: out.write(_in.read())


def normalize_paths(paths):
    """Normalize paths for cross-platform compatibility.
    
    This function normalizes file paths to handle differences between
    platforms (e.g., backslashes on Windows vs forward slashes on Unix).
    """
    return set(os.path.normpath(p) for p in paths)


from pytest import raises
def assert_raises_regex(expected, regexp, callable, *a, **kw):
    raises(expected, callable, *a, **kw).match(regexp)


try:
    from test.support import check_warnings  # Python 3
except ImportError:

    try:
        from test.test_support import check_warnings   # Python 2
    except ImportError:
        # Python < 2.6
        import contextlib

        @contextlib.contextmanager
        def check_warnings(*filters, **kwargs):
            # We cannot reasonably support this, we'd have to copy to much code.
            # (or write our own). Since this is only testing warnings output,
            # we might slide by ignoring it.
            yield



from webassets.filter import Filter


__all__ = ('RCSSMin',)


class RCSSMin(Filter):
    """Minifies CSS.

    Requires the ``rcssmin`` package (https://github.com/ndparker/rcssmin).
    Alike 'cssmin' it is a port of the YUI CSS compression algorithm but aiming
    for speed instead of maximum compression.

    Supported configuration options:
    RCSSMIN_KEEP_BANG_COMMENTS (boolean)
        Keep bang-comments (comments starting with an exclamation mark).
    """

    name = 'rcssmin'
    options = {
        'keep_bang_comments': 'RCSSMIN_KEEP_BANG_COMMENTS',
    }

    def setup(self):
        super(RCSSMin, self).setup()
        try:
            import rcssmin
        except ImportError:
            raise EnvironmentError('The "rcssmin" package is not installed.')
        else:
            self.rcssmin = rcssmin

    def output(self, _in, out, **kw):
        keep = self.keep_bang_comments or False
        out.write(self.rcssmin.cssmin(_in.read(), keep_bang_comments=keep))
